# Generated by Django 2.2.18 on 2021-03-06 16:00

import django.db.models.deletion
import django_jsonfield_backport.models
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('scrud_django', '0003_resourcetype_revision'),
    ]

    operations = [
        migrations.AddField(
            model_name='resourcetype',
            name='indexing_policy',
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name='resource_type_indexing_policy',
                to='scrud_django.Resource',
                verbose_name='indexing policy',
            ),
        ),
        migrations.AlterField(
            model_name='resource',
            name='content',
            field=django_jsonfield_backport.models.JSONField(),
        ),
        migrations.AlterField(
            model_name='resourcetype',
            name='context',
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name='resource_type_context',
                to='scrud_django.Resource',
                verbose_name='linked-data context',
            ),
        ),
        migrations.AlterField(
            model_name='resourcetype',
            name='schema',
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name='resource_type_schema',
                to='scrud_django.Resource',
                verbose_name='schema',
            ),
        ),
    ]
